import asyncio
import os
from abc import ABC, abstractmethod
from asyncio import Lock
from asyncio.subprocess import STDOUT
from functools import partial
from typing import Any, MutableMapping, MutableSequence, Optional, Tuple, Union

import cachetools
import pkg_resources
from cachetools import Cache, TTLCache

from streamflow.core import utils
from streamflow.core.asyncache import cachedmethod
from streamflow.core.context import StreamFlowContext
from streamflow.core.scheduling import Location
from streamflow.deployment.connector.ssh import SSHConnector
from streamflow.log_handler import logger


class QueueManagerConnector(SSHConnector, ABC):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 file: str,
                 hostname: str,
                 username: str,
                 checkHostKey: bool = True,
                 dataTransferConnection: Optional[Union[str, MutableMapping[str, Any]]] = None,
                 maxConcurrentJobs: Optional[int] = 1,
                 maxConcurrentSessions: Optional[int] = 10,
                 passwordFile: Optional[str] = None,
                 pollingInterval: int = 5,
                 sshKey: Optional[str] = None,
                 sshKeyPassphraseFile: Optional[str] = None,
                 transferBufferSize: int = 2 ** 16) -> None:
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            checkHostKey=checkHostKey,
            dataTransferConnection=dataTransferConnection,
            file=file,
            maxConcurrentSessions=maxConcurrentSessions,
            nodes=[hostname],
            passwordFile=passwordFile,
            sshKey=sshKey,
            sshKeyPassphraseFile=sshKeyPassphraseFile,
            transferBufferSize=transferBufferSize,
            username=username)
        self.hostname: str = hostname
        self.maxConcurrentJobs: int = maxConcurrentJobs
        self.pollingInterval: int = pollingInterval
        self.scheduledJobs: MutableSequence[str] = []
        self.jobsCache: Cache = TTLCache(maxsize=1, ttl=self.pollingInterval)
        self.jobsCacheLock: Lock = Lock()

    @abstractmethod
    async def _get_output(self,
                          job_id: str,
                          location: str) -> str:
        ...

    @abstractmethod
    async def _get_returncode(self,
                              job_id: str,
                              location: str) -> int:
        ...

    @abstractmethod
    async def _get_running_jobs(self,
                                location: str) -> bool:
        ...

    @abstractmethod
    async def _remove_jobs(self,
                           location: str) -> None:
        ...

    @abstractmethod
    async def _run_batch_command(self,
                                 helper_file: str,
                                 job_name: str,
                                 location: str,
                                 workdir: Optional[str] = None,
                                 stdin: Optional[Union[int, str]] = None,
                                 stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                                 stderr: Union[int, str] = asyncio.subprocess.STDOUT) -> str:
        ...

    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        return {self.hostname: Location(
            name=self.hostname,
            hostname=self.hostname,
            slots=self.maxConcurrentJobs)}

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'queue_manager.json'))

    async def run(self,
                  location: str,
                  command: MutableSequence[str],
                  environment: MutableMapping[str, str] = None,
                  workdir: Optional[str] = None,
                  stdin: Optional[Union[int, str]] = None,
                  stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                  stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                  capture_output: bool = False,
                  job_name: Optional[str] = None) -> Optional[Tuple[Optional[Any], int]]:
        # TODO: find a smarter way to identify detachable jobs when implementing stacked connectors
        if job_name:
            command = utils.create_command(
                command=command,
                environment=environment,
                workdir=workdir)
            logger.debug("Executing command {command} on {location} {job}".format(
                command=command,
                location=location,
                job="for job {job}".format(job=job_name) if job_name else ""))
            command = utils.encode_command(command)
            helper_file = await self._build_helper_file(command, location, environment, workdir)
            job_id = await self._run_batch_command(
                helper_file=helper_file,
                job_name=job_name,
                location=location,
                workdir=workdir,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr)
            logger.info("Scheduled job {job} with job id {job_id}".format(
                job=job_name,
                job_id=job_id))
            self.scheduledJobs.append(job_id)
            async with self.jobsCacheLock:
                self.jobsCache.clear()
            while True:
                async with self.jobsCacheLock:
                    running_jobs = await self._get_running_jobs(location)
                if job_id not in running_jobs:
                    break
                await asyncio.sleep(self.pollingInterval)
            self.scheduledJobs.remove(job_id)
            return (
                await self._get_output(job_id, location) if stdout == STDOUT else None,
                await self._get_returncode(job_id, location))
        else:
            return await super().run(
                location=location,
                command=command,
                environment=environment,
                workdir=workdir,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                job_name=job_name,
                capture_output=capture_output)

    async def undeploy(self, external: bool) -> None:
        await self._remove_jobs(self.hostname)
        self.scheduledJobs = {}
        await super().undeploy(external)


class SlurmConnector(QueueManagerConnector):

    async def _get_output(self,
                          job_id: str,
                          location: str) -> str:
        async with self._get_ssh_client(location) as ssh_client:
            output_path = (await ssh_client.run(
                "scontrol show -o job {job_id} | sed -n 's/^.*StdOut=\\([^[:space:]]*\\).*/\\1/p'".format(
                    job_id=job_id
                ))).stdout.strip()
            return ((await ssh_client.run('cat {output}'.format(output=output_path))).stdout.strip()
                    if output_path else "")

    async def _get_returncode(self,
                              job_id: str,
                              location: str) -> int:
        async with self._get_ssh_client(location) as ssh_client:
            return int((await ssh_client.run(
                "scontrol show -o job {job_id} | sed -n 's/^.*ExitCode=\\([0-9]\\+\\):.*/\\1/p'".format(
                    job_id=job_id
                ))).stdout.strip())

    @cachedmethod(lambda self: self.jobsCache, key=partial(cachetools.keys.hashkey, 'running_jobs'))
    async def _get_running_jobs(self,
                                location: str) -> MutableSequence[str]:
        async with self._get_ssh_client(location) as ssh_client:
            return [j.strip() for j in (await ssh_client.run(
                "squeue -h -j {job_ids} -t {states} -O JOBID".format(
                    job_ids=",".join(self.scheduledJobs),
                    states=",".join(['PENDING', 'RUNNING', 'SUSPENDED', 'COMPLETING', 'CONFIGURING',
                                     'RESIZING', 'REVOKED', 'SPECIAL_EXIT'])
                ))).stdout.strip().splitlines()]

    async def _remove_jobs(self,
                           location: str) -> None:
        async with self._get_ssh_client(location) as ssh_client:
            await ssh_client.run(
                "scancel {job_ids}".format(job_ids=" ".join(self.scheduledJobs)))

    async def _run_batch_command(self,
                                 helper_file: str,
                                 job_name: str,
                                 location: str,
                                 workdir: Optional[str] = None,
                                 stdin: Optional[Union[int, str]] = None,
                                 stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                                 stderr: Union[int, str] = asyncio.subprocess.STDOUT) -> str:
        batch_command = "sbatch --parsable {workdir} {stdin} {stdout} {stderr} {helper_file}".format(
            workdir="-D {workdir}".format(workdir=workdir) if workdir is not None else "",
            stdin="-i \"{stdin}\"".format(stdin=stdin) if stdin is not None else "",
            stdout="-o \"{stdout}\"".format(stdout=stdout) if stdout != STDOUT else "",
            stderr="-e \"{stderr}\"".format(stderr=stderr) if stderr != STDOUT and stderr != stdout else "",
            helper_file=helper_file)
        async with self._get_ssh_client(location) as ssh_client:
            result = await ssh_client.run(batch_command)
        return result.stdout.strip()


class PBSConnector(QueueManagerConnector):

    async def _get_output(self,
                          job_id: str,
                          location: str) -> str:
        async with self._get_ssh_client(location) as ssh_client:
            output_path = (await ssh_client.run(
                "qstat {job_id} -xf | sed -n 's/^\\s*Output_Path\\s=\\s.*:\\(.*\\)\\s*$/\\1/p'".format(
                    job_id=job_id
                ))).stdout.strip()
            return ((await ssh_client.run('cat {output}'.format(output=output_path))).stdout.strip()
                    if output_path else "")

    async def _get_returncode(self,
                              job_id: str,
                              location: str) -> int:
        async with self._get_ssh_client(location) as ssh_client:
            return int((await ssh_client.run(
                "qstat {job_id} -xf | sed -n 's/^\\s*Exit_status\\s=\\s\\([0-9]\\+\\)\\s*$/\\1/p'".format(
                    job_id=job_id
                ))).stdout.strip())

    @cachedmethod(lambda self: self.jobsCache, key=partial(cachetools.keys.hashkey, 'running_jobs'))
    async def _get_running_jobs(self,
                                location: str) -> MutableSequence[str]:
        async with self._get_ssh_client(location) as ssh_client:
            return (await ssh_client.run(
                "qstat -awx {job_ids} | grep '{grep_ids}' | awk '{{if($10 != \"E\" && $10 != \"F\") {{print $1}}}}'".format(
                    job_ids=" ".join(self.scheduledJobs),
                    grep_ids="\\|".join(self.scheduledJobs)
                ))).stdout.strip().splitlines()

    async def _remove_jobs(self,
                           location: str) -> None:
        async with self._get_ssh_client(location) as ssh_client:
            await ssh_client.run(
                "qdel {job_ids}".format(job_ids=" ".join(self.scheduledJobs)))

    async def _run_batch_command(self,
                                 helper_file: str,
                                 job_name: str,
                                 location: str,
                                 workdir: Optional[str] = None,
                                 stdin: Optional[Union[int, str]] = None,
                                 stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                                 stderr: Union[int, str] = asyncio.subprocess.STDOUT) -> str:
        batch_command = "{workdir} qsub {stdin} {stdout} {stderr} {helper_file}".format(
            workdir="cd {workdir} &&".format(workdir=workdir) if workdir is not None else "",
            stdin="-i \"{stdin}\"".format(stdin=stdin) if stdin is not None else "",
            stdout=("-o \"{stdout}\"".format(stdout=stdout if stdout != STDOUT else utils.random_name())),
            stderr="-e \"{stderr}\"".format(stderr=stderr) if stderr != STDOUT and stderr != stdout else "",
            helper_file=helper_file)
        async with self._get_ssh_client(location) as ssh_client:
            result = await ssh_client.run(batch_command)
        return result.stdout.strip()



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from enum import Enum
from typing import MutableSequence, TYPE_CHECKING

from streamflow.core.config import SchemaEntity

if TYPE_CHECKING:
    from streamflow.core.context import StreamFlowContext
    from typing import Any, Optional, Set

LOCAL_LOCATION = '__LOCAL__'


class DataType(Enum):
    PRIMARY = 0
    SYMBOLIC_LINK = 1
    INVALID = 2


class DataLocation(object):
    __slots__ = ('path', 'relpath', 'deployment', 'data_type', 'location', 'available')

    def __init__(self,
                 path: str,
                 relpath: str,
                 deployment: str,
                 data_type: DataType,
                 location: str,
                 available: bool = False):
        self.path: str = path
        self.relpath: str = relpath
        self.deployment: str = deployment
        self.location: str = location
        self.data_type: DataType = data_type
        self.available: asyncio.Event = asyncio.Event()
        if available:
            self.available.set()

    def __eq__(self, other):
        if not isinstance(other, DataLocation):
            return False
        else:
            return (self.path == other.path and
                    self.deployment == other.deployment and
                    self.location == other.location)

    def __hash__(self):
        return hash((self.path, self.deployment, self.location))


class DataManager(SchemaEntity):

    def __init__(self, context: StreamFlowContext):
        self.context: StreamFlowContext = context

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    def get_data_locations(self,
                           path: str,
                           deployment: Optional[str] = None,
                           location_type: Optional[DataType] = None) -> Set[DataLocation]:
        ...

    @abstractmethod
    def get_source_location(self,
                            path: str,
                            dst_deployment: str) -> Optional[DataLocation]:
        ...

    @abstractmethod
    def invalidate_location(self,
                            location: str,
                            path: str) -> None:
        ...

    @abstractmethod
    def register_path(self,
                      deployment: str,
                      location: str,
                      path: str,
                      relpath: str,
                      data_type: DataType = DataType.PRIMARY) -> DataLocation:
        ...

    @abstractmethod
    def register_relation(self,
                          src_location: DataLocation,
                          dst_location: DataLocation) -> None:
        ...

    @abstractmethod
    async def transfer_data(self,
                            src_deployment: str,
                            src_locations: MutableSequence[str],
                            src_path: str,
                            dst_deployment: str,
                            dst_locations: MutableSequence[str],
                            dst_path: str,
                            writable: bool = False):
        ...


class FileType(Enum):
    FILE = 1
    DIRECTORY = 2


class StreamWrapper(ABC):

    def __init__(self, stream):
        self.stream = stream
        self.closed = False

    @abstractmethod
    async def close(self): ...

    @abstractmethod
    async def read(self, size: Optional[int] = None): ...

    @abstractmethod
    async def write(self, data: Any): ...


class StreamWrapperContext(ABC):

    @abstractmethod
    async def __aenter__(self) -> StreamWrapper:
        ...

    @abstractmethod
    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        ...



from __future__ import annotations

import asyncio
import cgi
import errno
import glob
import os
import posixpath
from hashlib import sha1
from typing import MutableSequence, TYPE_CHECKING

import aiohttp

from streamflow.core import utils
from streamflow.core.context import StreamFlowContext
from streamflow.core.data import FileType
from streamflow.core.exception import WorkflowExecutionException
from streamflow.deployment.connector.local import LocalConnector

if TYPE_CHECKING:
    from streamflow.core.deployment import Connector
    from typing import Union, Optional


def _check_status(command: MutableSequence[str], location: str, result: str, status: int):
    if status != 0:
        raise WorkflowExecutionException("Command '{}' on location {} terminated with status {}: {}".format(
            ' '.join(command), location, status, result))


async def _file_checksum(context: StreamFlowContext,
                         connector: Connector,
                         location: Optional[str],
                         path: str) -> str:
    if isinstance(connector, LocalConnector):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            context.process_executor,
            _file_checksum_local,
            path)
    else:
        command = ["sha1sum {path} | awk '{{print $1}}'".format(path=path)]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        _check_status(command, location, result, status)
        return result.strip()


def _file_checksum_local(path: str) -> str:
    with open(path, "rb") as f:
        sha1_checksum = sha1()
        while data := f.read(2 ** 16):
            sha1_checksum.update(data)
        return sha1_checksum.hexdigest()


def _listdir_local(path: str, file_type: Optional[FileType]) -> MutableSequence[str]:
    content = []
    dir_content = os.listdir(path)
    check = (os.path.isfile if file_type == FileType.FILE else os.path.isdir) if file_type is not None else None
    for element in dir_content:
        element_path = os.path.join(path, element)
        if check and check(element_path):
            content.append(element_path)
    return content


async def checksum(context: StreamFlowContext,
                   connector: Connector,
                   location: Optional[str],
                   path: str) -> str:
    if await isfile(connector, location, path):
        return await _file_checksum(context, connector, location, path)
    else:
        raise Exception("Checksum for folders is not implemented yet")


async def download(
        connector: Connector,
        locations: Optional[MutableSequence[str]],
        url: str, parent_dir: str) -> str:
    await mkdir(connector, locations, parent_dir)
    if isinstance(connector, LocalConnector):
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status == 200:
                    _, params = cgi.parse_header(response.headers.get('Content-Disposition', ''))
                    filepath = os.path.join(parent_dir, params['filename'])
                    content = await response.read()
                else:
                    raise Exception
        with open(filepath, mode="wb") as f:
            f.write(content)
    else:
        async with aiohttp.ClientSession() as session:
            async with session.head(url, allow_redirects=True) as response:
                if response.status == 200:
                    _, params = cgi.parse_header(response.headers.get('Content-Disposition', ''))
                    filepath = posixpath.join(parent_dir, params['filename'])
        download_tasks = []
        for location in locations:
            download_tasks.append(asyncio.create_task(connector.run(
                location=location,
                command=["if [ command -v curl ]; curl -L -o \"{path}\"; else wget -P \"{dir}\" {url}; fi".format(
                    dir=parent_dir,
                    path=filepath,
                    url=url)])))
        await asyncio.gather(*download_tasks)
    return filepath


async def exists(connector: Connector, location: Optional[str], path: str) -> bool:
    if isinstance(connector, LocalConnector):
        return os.path.exists(path)
    else:
        command = ["test -e \"{path}\"".format(path=path)]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        if status > 1:
            raise WorkflowExecutionException("Command '{}' on location {} terminated with status {}: {}".format(
                command, location, status, result))
        else:
            return not status


async def follow_symlink(connector: Connector, location: Optional[str], path: str) -> str:
    if isinstance(connector, LocalConnector):
        return os.path.realpath(path)
    else:
        command = ["readlink -f \"{path}\"".format(path=path)]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        _check_status(command, location, result, status)
        return result.strip()


async def head(connector: Connector, location: Optional[str], path: str, num_bytes: int) -> str:
    if isinstance(connector, LocalConnector):
        with open(path, "rb") as f:
            return f.read(num_bytes).decode('utf-8')
    else:
        command = ["head", "-c", str(num_bytes), path]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        _check_status(command, location, result, status)
        return result.strip()


async def isdir(connector: Connector, location: Optional[str], path: str) -> bool:
    if isinstance(connector, LocalConnector):
        return os.path.isdir(path)
    else:
        command = ["test -d \"{path}\"".format(path=path)]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        if status > 1:
            raise WorkflowExecutionException("Command '{}' on location {} terminated with status {}: {}".format(
                command, location, status, result))
        else:
            return not status


async def isfile(connector: Connector, location: Optional[str], path: str) -> bool:
    if isinstance(connector, LocalConnector):
        return os.path.isfile(path)
    else:
        command = ["test -f \"{path}\"".format(path=path)]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        if status > 1:
            raise WorkflowExecutionException("Command '{}' on location {} terminated with status {}: {}".format(
                command, location, status, result))
        else:
            return not status


async def listdir(connector: Connector,
                  location: Optional[str],
                  path: str,
                  file_type: Optional[FileType] = None) -> MutableSequence[str]:
    if isinstance(connector, LocalConnector):
        return _listdir_local(path, file_type)
    else:
        command = "find -L \"{path}\" -mindepth 1 -maxdepth 1 {type}".format(
            path=path,
            type="-type {type}".format(
                type="d" if file_type == FileType.DIRECTORY else "f") if file_type is not None else ""
        ).split()
        content, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        _check_status(command, location, content, status)
        content = content.strip(' \n')
        return content.split('\n') if content else []


async def mkdir(
        connector: Connector,
        locations: Optional[MutableSequence[str]],
        path: str) -> None:
    return await mkdirs(connector, locations, [path])


async def mkdirs(
        connector: Connector,
        locations: Optional[MutableSequence[str]],
        paths: MutableSequence[str]) -> None:
    if isinstance(connector, LocalConnector):
        for path in paths:
            os.makedirs(path, exist_ok=True)
    else:
        command = ["mkdir", "-p"]
        command.extend(paths)
        await asyncio.gather(*(asyncio.create_task(
            connector.run(location=location, command=command)
        ) for location in locations))


async def read(connector: Connector, location: Optional[str], path: str) -> str:
    if isinstance(connector, LocalConnector):
        with open(path, "rb") as f:
            return f.read().decode('utf-8')
    else:
        command = ["cat", path]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        _check_status(command, location, result, status)
        return result.strip()


async def resolve(
        connector: Connector,
        location: Optional[str],
        pattern: str) -> Optional[MutableSequence[str]]:
    if isinstance(connector, LocalConnector):
        return sorted(glob.glob(pattern))
    else:
        command = ["printf", "\"%s\\0\"", pattern, "|", "xargs", "-0", "-I{}",
                   "sh", "-c", "\"if [ -e \\\"{}\\\" ]; then echo \\\"{}\\\"; fi\"", "|", "sort"]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        _check_status(command, location, result, status)
        return result.split()


async def rm(
        connector: Connector,
        location: Optional[str],
        path: Union[str, MutableSequence[str]]) -> None:
    if isinstance(connector, LocalConnector):
        if isinstance(path, MutableSequence):
            for p in path:
                if os.path.exists(p):
                    os.remove(p)
        else:
            if os.path.exists(path):
                return os.remove(path)
    else:
        if isinstance(path, MutableSequence):
            path = ' '.join(["\"{path}\"".format(path=p) for p in path])
        else:
            path = "\"{path}\"".format(path=path)
        await connector.run(
            location=location,
            command=[''.join(["rm -rf ", path])])


async def size(
        connector: Connector,
        location: Optional[str],
        path: Union[str, MutableSequence[str]]) -> int:
    if not path:
        return 0
    elif isinstance(connector, LocalConnector):
        if isinstance(path, MutableSequence):
            weight = 0
            for p in path:
                weight += utils.get_size(p)
            return weight
        else:
            return utils.get_size(path)
    else:
        if isinstance(path, MutableSequence):
            path = ' '.join(["\"{path}\"".format(path=p) for p in path])
        else:
            path = "\"{path}\"".format(path=path)
        command = [''.join(["find -L ", path, " -type f -exec ls -ln {} \\+ | ",
                            "awk 'BEGIN {sum=0} {sum+=$5} END {print sum}'; "])]
        result, status = await connector.run(
            location=location,
            command=command,
            capture_output=True)
        _check_status(command, location, result, status)
        result = result.strip().strip("'\"")
        return int(result) if result.isdigit() else 0


async def symlink(connector: Connector, location: Optional[str], src: str, path: str) -> None:
    if isinstance(connector, LocalConnector):
        try:
            os.symlink(os.path.abspath(src), path, target_is_directory=os.path.isdir(path))
        except OSError as e:
            if not e.errno == errno.EEXIST:
                raise
    else:
        await connector.run(location=location, command=["ln", "-snf", src, path])


async def write(connector: Connector, location: Optional[str], path: str, content: str) -> None:
    if isinstance(connector, LocalConnector):
        with open(path, "w") as f:
            f.write(content)
    else:
        command = ["printf", "\"{content}\"".format(content=content.replace('"', '\\"'))]
        result, status = await connector.run(
            location=location,
            command=command,
            stdout=path,
            capture_output=True)
        _check_status(command, location, result, status)



import argparse
import os
import uuid

parser = argparse.ArgumentParser(description='StreamFlow Command Line')
subparsers = parser.add_subparsers(dest='context')

# streamflow list
list_parser = subparsers.add_parser('list')
list_parser.add_argument('streamflow_file', nargs='?', metavar='STREAMFLOW_FILE', default='streamflow.yml', type=str,
                         help='Path to the StreamFlow file describing the workflow execution.')
list_parser.add_argument('--outdir', default=os.getcwd(), type=str,
                         help='Output directory in which the final results of the workflow are stored.')

# streamflow run
run_parser = subparsers.add_parser('run')
run_parser.add_argument('streamflow_file', metavar='STREAMFLOW_FILE', default='streamflow.yml', type=str,
                        help='Path to the StreamFlow file describing the workflow execution.')
run_parser.add_argument('--name', default=str(uuid.uuid4()), type=str,
                        help='Name of the current workflow. Will be used for search and indexing.')
run_parser.add_argument('--outdir', default=os.getcwd(), type=str,
                        help='Output directory in which to store final results of the workflow.')
run_parser.add_argument('--quiet', action='store_true',
                        help='Only prints results, warnings and errors.')
run_parser.add_argument('--debug', action='store_true',
                        help='Prints debug-level diagnostic output.')

# streamflow report
report_parser = subparsers.add_parser('report')
report_parser.add_argument('workflow', metavar='WORKFLOW', type=str,
                           help='Name of the workflow to process.')
report_parser.add_argument('streamflow_file', nargs='?', metavar='STREAMFLOW_FILE',
                           default='streamflow.yml', type=str,
                           help='Path to the StreamFlow file describing the workflow execution.')
report_parser.add_argument('--group-by-step', action='store_true',
                           help='Groups execution of the same steps on a single line.')
report_parser.add_argument('--format', default='html', type=str,
                           help='Report format: html (default), pdf, eps, png, jpg, webp, svg, csv, or json.')
report_parser.add_argument('--name', type=str,
                           help='Name of the report file (default report.format).')
report_parser.add_argument('--outdir', default=os.getcwd(), type=str,
                           help='Output directory in which the final results of the workflow are stored.')

# streamflow version
version_parser = subparsers.add_parser('version', help='Only print StreamFlow version and exit.')



from streamflow.deployment.connector.container import DockerComposeConnector, DockerConnector, SingularityConnector
from streamflow.deployment.connector.kubernetes import Helm3Connector
from streamflow.deployment.connector.local import LocalConnector
from streamflow.deployment.connector.occam import OccamConnector
from streamflow.deployment.connector.queue_manager import PBSConnector, SlurmConnector
from streamflow.deployment.connector.ssh import SSHConnector

connector_classes = {
    'docker': DockerConnector,
    'docker-compose': DockerComposeConnector,
    'helm': Helm3Connector,
    'helm3': Helm3Connector,
    'local': LocalConnector,
    'occam': OccamConnector,
    'pbs': PBSConnector,
    'singularity': SingularityConnector,
    'slurm': SlurmConnector,
    'ssh': SSHConnector
}



import asyncio
import os
import posixpath
import re
from typing import Any, MutableMapping, MutableSequence, Optional, Tuple, Union

import asyncssh
import pkg_resources
from ruamel.yaml import YAML

from streamflow.core import utils
from streamflow.core.context import StreamFlowContext
from streamflow.core.scheduling import Location
from streamflow.deployment.connector.ssh import SSHConnector
from streamflow.log_handler import logger


class OccamConnector(SSHConnector):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 file: str,
                 sshKey: str,
                 username: str,
                 sshKeyPassphraseFile: Optional[str] = None,
                 hostname: Optional[str] = "occam.c3s.unito.it",
                 transferBufferSize: int = 2 ** 16) -> None:
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            nodes=[hostname],
            sshKey=sshKey,
            sshKeyPassphraseFile=sshKeyPassphraseFile,
            sharedPaths=[
                '/archive/home/{username}'.format(username=username),
                '/scratch/home/{username}'.format(username=username)],
            transferBufferSize=transferBufferSize,
            username=username)
        with open(os.path.join(context.config_dir, file)) as f:
            yaml = YAML(typ='safe')
            self.env_description = yaml.load(f)
        self.jobs_table: MutableMapping[str, MutableSequence[str]] = {}

    def _get_effective_locations(self,
                                 locations: MutableSequence[str],
                                 dest_path: str,
                                 source_location: Optional[str] = None) -> MutableSequence[str]:
        # If destination path is in a shared location, transfer only on the first location
        for shared_path in self.sharedPaths:
            if dest_path.startswith(shared_path):
                if source_location in locations:
                    return [source_location]
                else:
                    return [locations[0]]
        # Otherwise, check if some locations share bind mounts to the same shared local path
        common_paths = {}
        effective_locations = []
        for location in locations:
            shared_path = self._get_shared_path(location, dest_path)
            if shared_path is not None:
                if shared_path not in common_paths:
                    common_paths[shared_path] = location
                    effective_locations.append(location)
                elif location == source_location:
                    effective_locations.remove(common_paths[shared_path])
                    common_paths[shared_path] = location
                    effective_locations.append(location)
            else:
                effective_locations.append(location)

        return effective_locations

    def _get_shared_path(self, location: str, path: str) -> Optional[str]:
        for shared_path in self.sharedPaths:
            if path.startswith(shared_path):
                return path
        volumes = self._get_volumes(location)
        for volume in volumes:
            local, remote = volume.split(':')
            if path.startswith(remote):
                return posixpath.normpath(posixpath.join(local, posixpath.relpath(path, remote)))
        return None

    async def _get_tmpdir(self, location: str):
        scratch_home = '/scratch/home/{username}'.format(username=self.username)
        temp_dir = posixpath.join(scratch_home, 'streamflow', "".join(utils.random_name()))
        async with self._get_ssh_client(location) as ssh_client:
            await ssh_client.run('mkdir -p {dir}'.format(dir=temp_dir))
        return temp_dir

    def _get_volumes(self, location: str) -> MutableSequence[str]:
        for name in self.jobs_table:
            if location in self.jobs_table[name]:
                service = name
                return self.env_description[service].get('volumes', [])

    async def _copy_remote_to_remote(self,
                                     src: str,
                                     dst: str,
                                     locations: MutableSequence[str],
                                     source_location: str,
                                     read_only: bool = False) -> None:
        effective_locations = self._get_effective_locations(locations, dst, source_location)
        # Check for the need of a temporary copy
        temp_dir = None
        for location in effective_locations:
            if source_location != location:
                temp_dir = await self._get_tmpdir(location)
                copy1_command = ['/bin/cp', '-rf', src, temp_dir]
                await self.run(source_location, copy1_command)
                break
        # Perform the actual copies
        await asyncio.gather(*(asyncio.create_task(
            self._copy_remote_to_remote_single(
                src=src,
                dst=dst,
                location=location,
                source_location=source_location,
                temp_dir=temp_dir,
                read_only=read_only)
        ) for location in effective_locations))
        # If a temporary location was created, delete it
        if temp_dir is not None:
            for location in effective_locations:
                async with self._get_ssh_client(location) as ssh_client:
                    await ssh_client.run('rm -rf {dir}'.format(dir=temp_dir))

    async def _copy_remote_to_remote_single(self,
                                            src: str,
                                            dst: str,
                                            location: str,
                                            source_location: str,
                                            temp_dir: Optional[str],
                                            read_only: bool = False) -> None:
        if source_location == location:
            command = ['/bin/cp', "-rf", src, dst]
            await self.run(location, command)
        else:
            copy2_command = ['/bin/cp', '-rf', temp_dir + "/*", dst]
            await self.run(location, copy2_command)

    async def _copy_local_to_remote(self,
                                    src: str,
                                    dst: str,
                                    locations: MutableSequence[str],
                                    read_only: bool = False) -> None:
        effective_locations = self._get_effective_locations(locations, dst)
        # Check for the need of a temporary copy
        temp_dir = None
        for location in effective_locations:
            shared_path = self._get_shared_path(location, dst)
            if shared_path is None:
                temp_dir = await self._get_tmpdir(location)
                async with self._get_ssh_client(location) as ssh_client:
                    await asyncssh.scp(src, (ssh_client, temp_dir), preserve=True, recurse=True)
                break
        # Perform the actual copies
        copy_tasks = []
        for location in effective_locations:
            copy_tasks.append(asyncio.create_task(
                self._copy_local_to_remote_single(
                    src=src,
                    dst=dst,
                    location=location,
                    temp_dir=temp_dir,
                    read_only=read_only)))
        await asyncio.gather(*copy_tasks)
        # If a temporary location was created, delete it
        if temp_dir is not None:
            for location in effective_locations:
                async with self._get_ssh_client(location) as ssh_client:
                    await ssh_client.run('rm -rf {dir}'.format(dir=temp_dir))

    async def _copy_local_to_remote_single(self,
                                           src: str,
                                           dst: str,
                                           location: str,
                                           temp_dir: Optional[str],
                                           read_only: bool = False) -> None:
        shared_path = self._get_shared_path(location, dst)
        if shared_path is not None:
            async with self._get_ssh_client(location) as ssh_client:
                await asyncssh.scp(src, (ssh_client, shared_path), preserve=True, recurse=True)
        else:
            copy_command = ['/bin/cp', "-rf", temp_dir + "/*", dst]
            await self.run(location, copy_command)

    async def _copy_remote_to_local(self,
                                    src: str,
                                    dst: str,
                                    location: str,
                                    read_only: bool = False) -> None:
        shared_path = self._get_shared_path(location, src)
        if shared_path is not None:
            async with self._get_ssh_client(location) as ssh_client:
                await asyncssh.scp((ssh_client, shared_path), dst, preserve=True, recurse=True)
        else:
            temp_dir = await self._get_tmpdir(location)
            copy_command = ['/bin/cp', "-rf", src, temp_dir, '&&',
                            'find', temp_dir, '-maxdepth', '1', '-mindepth', '1', '-exec', 'basename', '{}', '\\;']
            contents, _ = await self.run(location, copy_command, capture_output=True)
            contents = contents.split()
            scp_tasks = []
            async with self._get_ssh_client(location) as ssh_client:
                for content in contents:
                    scp_tasks.append(asyncio.create_task(asyncssh.scp(
                        (ssh_client, posixpath.join(temp_dir, content)),
                        dst,
                        preserve=True,
                        recurse=True
                    )))
                await asyncio.gather(*scp_tasks)
            delete_command = ['rm', '-rf', temp_dir]
            await self.run(location, delete_command)

    async def _deploy_node(self, name: str, service: MutableMapping[str, Any], node: str):
        deploy_command = "".join([
            "{workdir}"
            "occam-run ",
            "{x11}",
            "{node}",
            "{stdin}"
            "{jobidFile}"
            "{shmSize}"
            "{volumes}"
            "{image} "
            "{command}"
        ]).format(
            workdir="cd {workdir} && ".format(workdir=service.get('workdir')) if 'workdir' in service else "",
            x11=self.get_option("x", service.get('x11')),
            node=self.get_option("n", node),
            stdin=self.get_option("i", service.get('stdin')),
            jobidFile=self.get_option("c", service.get('jobidFile')),
            shmSize=self.get_option("s", service.get('shmSize')),
            volumes=self.get_option("v", service.get('volumes')),
            image=service['image'],
            command=" ".join(service.get('command', ""))
        )
        logger.debug("Executing {command}".format(command=deploy_command))
        async with self._get_ssh_client(name) as ssh_client:
            result = await ssh_client.run(deploy_command)
        output = result.stdout
        search_result = re.findall('({node}-[0-9]+).*'.format(node=node), output, re.MULTILINE)
        if search_result:
            if name not in self.jobs_table:
                self.jobs_table[name] = []
            self.jobs_table[name].append(search_result[0])
            logger.info("Deployed {name} on {location}".format(name=name, location=search_result[0]))
        else:
            raise Exception

    async def _undeploy_node(self, name: str, job_id: str):
        undeploy_command = "".join([
            "occam-kill ",
            "{job_id}"
        ]).format(
            job_id=job_id
        )
        logger.debug("Executing {command}".format(command=undeploy_command))
        async with self._get_ssh_client(name) as ssh_client:
            await ssh_client.run(undeploy_command)
        logger.info("Killed {location}".format(location=job_id))

    async def deploy(self, external: bool) -> None:
        await super().deploy(external)
        if not external:
            deploy_tasks = []
            for (name, service) in self.env_description.items():
                nodes = service.get('nodes', ['node22'])
                for node in nodes:
                    deploy_tasks.append(asyncio.create_task(self._deploy_node(name, service, node)))
            await asyncio.gather(*deploy_tasks)

    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        nodes = self.jobs_table.get(service, []) if service else utils.flatten_list(self.jobs_table.values())
        return {n: Location(name=n, hostname=n.split('-')[0]) for n in nodes}

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'occam.json'))

    async def undeploy(self, external: bool) -> None:
        if not external:
            # Undeploy
            undeploy_tasks = []
            for name in self.jobs_table:
                for job_id in self.jobs_table[name]:
                    undeploy_tasks.append(asyncio.create_task(self._undeploy_node(name, job_id)))
            await asyncio.gather(*undeploy_tasks)
            self.jobs_table = {}
        # Close connection
        await super().undeploy(external)

    async def run(self,
                  location: str,
                  command: MutableSequence[str],
                  environment: MutableMapping[str, str] = None,
                  workdir: Optional[str] = None,
                  stdin: Optional[Union[int, str]] = None,
                  stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                  stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                  capture_output: bool = False,
                  job_name: Optional[str] = None) -> Optional[Tuple[Optional[Any], int]]:
        command = self._get_command(
            location=location,
            command=command,
            environment=environment,
            workdir=workdir,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            job_name=job_name)
        occam_command = "".join(
            "occam-exec "
            "{location} "
            "sh -c '{command}'"
        ).format(
            location=location,
            command=command)
        async with self._get_ssh_client(location) as ssh_client:
            result = await ssh_client.run(occam_command)
        if capture_output:
            lines = (line for line in result.stdout.split('\n'))
            out = ""
            for line in lines:
                if line.startswith("Trying to exec commands into container"):
                    break
            try:
                line = next(lines)
                out = line.strip(' \r\t')
            except StopIteration:
                return out
            for line in lines:
                out = "\n".join([out, line.strip(' \r\t')])
            return out, result.returncode
